const { _, log } = require('./basd')
// const Crypto = require('basd/crypto')

class SearchInterpreter {
  constructor(subject) {
    this.subject = subject
  }
  run(nodes) {
    return this.visitNodes(nodes)
  }
  visitNodes(nodes) {
    if (!nodes.left)
      return this.visitNode(nodes)
    const leftNode = this.visitNode(nodes.left)
    if (!nodes.right)
      return leftNode
    const operator = nodes.operator
    const rightNode = this.visitNode(nodes.right)
    if (['&', 'and', 'AND'].includes(operator)) {
      return leftNode && rightNode
    } else if (['|', 'or', 'OR', '<implicit>'].includes(operator)) {
      return leftNode || rightNode
    }
  }
  visitNode(node) {
    if (node.left)
      return this.visitNodes(node)
    switch (node.type) {
      case 'word':
        return this.visitWord(node)
      case 'phrase':
        return this.visitPhrase(node)
    }
  }
  visitWord(node) {
    return this.subject.indexOf(node.text) > -1
  }
  visitPhrase(node) {
    return this.subject.indexOf(node.text.replace(/"/g, '')) > -1
  }
}

class SearchQuery {
  static get Interpreter() { return SearchInterpreter }
  constructor(opts = {}) {
    this.opts = _.defaults(opts, {})
    this.types = {
      GROUP: 'group',
      PHRASE: 'phrase',
      OPERATOR: 'operator',
      WORD: 'word',
      WHITESPACE: 'whitespace',
      UNKNOWN: 'unknown'
    }
    this.pipeline = [
      this.types.GROUP,
      this.types.PHRASE,
      this.types.OPERATOR,
      this.types.WORD,
      this.types.WHITESPACE
    ]
    this.patterns = {
      [this.types.GROUP]: /^\([^)]+\)/,
      [this.types.PHRASE]: /^("[^"]+")/,
      [this.types.OPERATOR]: opts.operators || /^(AND|OR|NOT|and|or|not|\||&)/,
      // [this.types.OPERATOR]: /^(AND|OR|NOT)/, // <= orig
      [this.types.WORD]: /^([^\s]+)/,
      [this.types.WHITESPACE]: /^(\s+)/
    }
    this.tokenizers = {
      [this.types.PHRASE]: this.capture(this.patterns[this.types.PHRASE]),
      [this.types.GROUP]: (str) => this.snip(str),
      [this.types.OPERATOR]: this.capture(this.patterns[this.types.OPERATOR]),
      [this.types.WORD]: this.capture(this.patterns[this.types.WORD]),
      [this.types.WHITESPACE]: this.capture(this.patterns[this.types.WHITESPACE]),
      [this.types.UNKNOWN]: () => '�'
    }
    this.ignore = new Set([ this.types.WHITESPACE ])
    this.negate = new Set([ this.types.OPERATOR ])
  }
  snip(str) {
    // snip-group
    let ii = 0
    let delimiters = 0
    while (ii < str.length) {
      const char = str.charAt(ii)
      ii++
      if (char === '(') {
        delimiters++
      } else if (char === ')') {
        delimiters--
        if (delimiters === 0) {
          break
        }
      }
    }
    return str.substr(0, ii)
  }
  unwrap(str) {
    // unwrap-group
    return str.replace(/^\(/, '').replace(/\)$/, '')
  }
  capture(regexp) {
    // capture-match
    return str => regexp.exec(str).pop()
  }
  tokenize(query) {
    if (!_.isStr(query))
      throw new Error(`the query to parse must be a string`)
    const tokens = []
    let ii = 0
    while (ii < query.length) {
      const substr = query.substring(ii)
      const type = this.pipeline.find(type => this.patterns[type].test(substr)) || this.types.UNKNOWN
      const text = this.tokenizers[type](substr)
      const token = { type, text, offset: ii, length: text.length }
      if (type === this.types.GROUP)
        token.children = this.tokenize(this.unwrap(text))
      if (this.negate.has(type)) {
        const last = tokens[tokens.length - 1]
        if (last && last.type === type)
          tokens.pop()
      }
      if (!this.ignore.has(type))
        tokens.push(token)
      ii += token.length
    }
    return tokens
  }
  getWords(tokens) {
    let words = []
    for (let token of tokens) {
      if (token.type === 'word') {
        words.push(token.text)
      } else if (token.type === 'group') {
        words = words.concat(this.getWords(token.children))
      } else if (token.type === 'phrase') {
        let _words = token.text.replace(/"/g, '').split(/\s+/)
        for (let _word of _words)
          words.push(_word)
      }
    }
    return _.uniq(words.map(word => _.lc(word)))
  }
  builder(tokens) {
    const tree = { left: null, operator: '<implicit>', right: null }
    const lhs = tokens[0]
    const rhs = tokens[1]
    let flag
    let group
    if (lhs.type === this.types.WORD || lhs.type === this.types.PHRASE)
      tree.left = lhs
    if (lhs.type === this.types.GROUP)
      group = tree.left = this.builder(lhs.children)
    if (rhs && rhs.type === this.types.OPERATOR)
      flag = tree.operator = rhs.text
    const next = tokens.slice(flag ? 2 : 1)
    if (next.length)
      tree.right = this.builder(next)
    else
      return group ? { left: tree.left } : tree.left
    return tree
  }
  parse(query) {
    const tokens = this.tokenize(query)
    return this.builder(tokens)
  }
  interpret(ast, subject) {
    const interpreter = new SearchInterpreter(subject)
    return interpreter.run(ast)
  }
  filter(query) {
    const ast = this.parse(query)
    // _.print(ast)
    return subject => this.interpret(ast, subject)
  }
  static filter(query, opts) {
    const sq = new this(opts)
    return sq.filter(query)
  }
  static getWords(term) {
    const query = new this()
    const tokens = query.tokenize(term)
    return query.getWords(tokens)
  }
  static parseTerm(term, black) {
    black = black || ['and', 'or', 'the']
    // term = "This., -/ is #! an $ % ^ & * example ;: {} of a = -_ [string] /\\ with `~)() punctuation"
    return term
      .replace(/[\/\[\]\^\*\-\\"?&.,#!$%;:'`{}=_~()]/g, '')
      .replace(/\s{2,}/g,' ')
      .split(' ')
      .map(word => _.lc(word))
      .filter(word => !black.includes(word))
  }
  static parseWords(term, id, dict = {}, black) {
    const words = this.parseTerm(term, black)
    id = id || _.hash(words.join(''))
    for (const word of words) {
      if (!dict[word])
        dict[word] = { len: 0, refs: [] }
      const node = dict[word]
      node.len++
      node.refs.push(id)
      node.refs = _.uniq(node.refs)
    }
    return dict
  }

  /* // might not use -- don't want crypto dep
  static parseWords(term, id, dict = {}, black) {
    const words = this.parseTerm(term, black)
    id = id || Crypto.hash({ words: words.join('') }, '10')
    for (const word of words) {
      if (!dict[word])
        dict[word] = { len: 0, refs: [] }
      const node = dict[word]
      node.len++
      node.refs.push(id)
      node.refs = _.uniq(node.refs)
    }
    return dict
  }
  */
  static searchV1(term, dict = {}) {
    return _(this.parseTerm(term))
      .transform((acc, cur) => dict[cur] ? acc.push(dict[cur].refs) : acc, [])
      .flatten()
      .uniq()
      .value()
  }
}

class Query {
  static get Query() { return Query }
  static get Builder() { return QueryBuilder }
  static get QueryBuilder() { return QueryBuilder }
  static get Search() { return SearchQuery }
  static get SearchQuery() { return SearchQuery }
  static checkType(val, typeId) {
    if (typeof typeId === 'function')
      typeId = typeId.name.toLowerCase()
    switch (typeId) {
      case 'number':
        return typeof val === 'number'
      case 'string':
        return typeof val === 'string'
      case 'object':
        return typeof val === 'object' && !(val instanceof Array)
      case 'array':
        return val instanceof Array
      case 'boolean':
        return typeof val === 'boolean'
      case 'null':
        return typeof val === 'object' && !val
      default:
        return false
    }
  }
  static checkWhere(obj, expression) {
    if (_.isStr(expression))
      return Reflect.construct(Function, [`return ${expression}`]).apply(obj)
    if (!_.isFn(expression))
      return false
    return expression.apply(obj)
  }
  static valOp(op, val, args, haystack) {
    let matchCount, matches = 0
    switch (op) {
      case '$eq':
        return this.eq(val, args)
      case '$in':
        for (let ii = 0; ii < args.length; ii++)
          if (this.eq(val, args[ii]))
            matches++
        return matches > 0
      case '$nin':
        return !this.valOp('$in', val, args, haystack)
      case '$all':
        matchCount = args.length
        if (val) {
          for (let ii = 0; ii < args.length; ii++) {
            for (jj = 0; jj < val.length; jj++) {
              if (this.eq(val[jj], args[ii])) {
                matches++
                break
              }
            }
          }
        }
        return matches === matchCount;
      case '$elemMatch':
        return this.match(args, val)
      case '$or':
        for (let ii = 0; ii < args.length; ii++)
          if (this.valOpMatch(val, args[ii], haystack))
            matches++
        return matches > 0
      case '$and':
        matchCount = args.length
        for (let ii = 0; ii < args.length; ii++)
          if (this.valOpMatch(val, args[ii], haystack))
            matches++
        return matches === matchCount
      case '$not':
        return !this.valOpMatch(val, args, haystack)
      case '$gt':
        return val > args
      case '$gte':
        return val >= args
      case '$ne':
        return val !== args
      case '$lt':
        return val < args
      case '$lte':
        return val <= args
      case '$mod':
        return (val % args[0]) === args[1]
      case '$size':
        return val instanceof Array && val.length === args
      case '$exists':
        if (args)
          return !_.isNil(val)
        else
          return _.isNil(val)
        // if (args)
        //   return val !== void 0
        // else
        //   return val === void 0
        break
      case '$contains':
        return _.lc(val).indexOf(args) !== -1
      case '$type':
        return this.checkType(val, args)
      case '$where':
        return this.checkWhere(val, args)
      case '$search':
        return SearchQuery.filter(args)(val)
      default:
        return false
    }
  }
  static valOpMatch(value, predicate, haystack) {
    const matchCount = Object.keys(predicate).length
    let matches = 0
    for (const needle in predicate) {
      const val = predicate[needle]
      if (needle[0] === '$')
        if (this.valOp(needle, value, val, haystack))
          matches++
    }
    return matches === matchCount
  }
  static operator(op, predicate, haystack) {
    let matches = 0
    switch (op) {
      case '$or':
        for (let ii = 0; ii < predicate.length; ii++)
          if (this.match(predicate[ii], haystack))
            matches++
        return matches > 0
      case '$and':
        const matchCount = predicate.length
        for (let ii = 0; ii < predicate.length; ii++)
          if (this.match(predicate[ii], haystack))
            matches++
        return matches === matchCount
      case '$not':
        return !this.match(predicate, haystack)
      case '$where':
        return this.checkWhere(haystack, predicate)
      default:
        return false
    }
  }
  static eq(a, b) {
    if (b instanceof RegExp) {
      if (Array.isArray(a)) {
        for (let ii = 0; ii < a.length; ii++) {
          const val = a[ii]
          if (b.test(val))
            return true
        }
        return false
      } else {
        return b.test(a)
      }
    } else if (Array.isArray(a)) {
      return a.indexOf(b) !== -1
    } else {
      return b == a
      // return b === a // use 2 for string numbers
    }
  }
  static lookup(needle, haystack) {
    const keys = needle.split('.')
    let val = haystack
    for (let ii = 0; ii < keys.length; ii++) {
      const key = keys[ii]
      if (val === void 0) // undefined
        return
      val = val[key]
    }
    return val
  }
  static get keys() {
    return ['limit', 'sort', 'skip', 'select']
  }
  static getOpts(query = {}, opts = {}) {
    const ignore = []
    for (const key of this.keys) {
      if (_.isNil(query[`$${key}`]))
        continue
      opts[key] = query[`$${key}`]
      ignore.push(`$${key}`)
    }
    return { ignore, opts }
  }
  static setOpts(query = {}, opts = {}) {
    const { ignore } = this.getOpts(query, opts)
    return _.omit(query, ignore)
  }

  static hasArr(needle, haystack) {
    let keys = needle.split('.')
    let val = haystack
    let path = []
    if (_.isArr(val))
      return path
    for (let ii = 0; ii < keys.length; ii++) {
      const key = keys[ii]
      if (_.isArr(val)) {
        return path
      }
      if (!_.isObj(val))
        return path
      path.push(key)
      val = val[key]
    }
    return []
  }

  static matchNested(needle, val, haystack) {
    let path = this.hasArr(needle, haystack)
    if (path.length) {
      let arr = _.get(haystack, path.join('.'))
      let ndl = _.difference(needle.split('.'), path).join('.')
      for (let item of arr) {
        let _path = this.hasArr(ndl, item)
        if (_path.length)
          return this.matchNested(_path.join('.'), val, this.lookup(ndl, item))
        let obj = this.lookup(ndl, item)
        if (_.isObj(val)) {
          if (this.valOpMatch(obj, val, haystack)) {
            return true
          }
        } else {
          if (this.eq(obj, val))
            return true
        }
      }
    } else {
      let obj = this.lookup(needle, haystack)
      if (_.isObj(val)) {
        if (this.valOpMatch(obj, val, haystack)) {
          return true
        }
      } else {
        if (this.eq(obj, val))
          return true
      }
    }
    return false
  }
  static match(predicate, haystack) {
    let matches = 0
    const matchCount = _.keys(predicate).length
    for (let needle in predicate) {
      const val = predicate[needle]
      if (needle[0] === '$') {
        matches += this.operator(needle, val, haystack)
      } else if (val && (val.constructor === Object || (typeof val === 'object' && !(val instanceof RegExp)))) {
        if (this.matchNested(needle, val, haystack))
          matches++
      } else {
        if (this.matchNested(needle, val, haystack))
          matches++
      }
    }
    return matches === matchCount
  }
  static matchV2(predicate, haystack) {
    let matches = 0
    const matchCount = _.keys(predicate).length
    for (let needle in predicate) {
      const val = predicate[needle]
      if (needle[0] === '$') {
        matches += this.operator(needle, val, haystack)
      } else if (val && (val.constructor === Object || (typeof val === 'object' && !(val instanceof RegExp)))) {

        // figure out if needle contains an arr, if so, loop through array with valOpMatch
        let goat = this.hasArr(needle, haystack)
        if (goat.length) {
          let arr = _.get(haystack, goat.join('.'))
          let ndl = _.difference(needle.split('.'), goat).join('.')
          for (let item of arr) {
            let goose = this.hasArr(ndl, item)
            if (!goose.length) {
              let o = this.lookup(ndl, item)
              if (this.valOpMatch(this.lookup(ndl, item), val, haystack)) {
                // matches++
                log('yay')
              }
            } else {
              // recurse...
            }
          }
        } else {
          if (this.valOpMatch(this.lookup(needle, haystack), val, haystack)) {
            matches++
          }
        }
        // xper ^

        // if (this.valOpMatch(this.lookup(needle, haystack), val, haystack)) {
        //   matches++
        // }
      } else {
        if (this.eq(this.lookup(needle, haystack), val))
          matches++
      }
    }
    return matches === matchCount
  }
  static matchV1(predicate, haystack) {
    let matches = 0
    const matchCount = _.keys(predicate).length
    for (let needle in predicate) {
      const val = predicate[needle]
      if (needle[0] === '$') {
        matches += this.operator(needle, val, haystack)
      } else if (val && (val.constructor === Object || (typeof val === 'object' && !(val instanceof RegExp)))) {
        if (this.valOpMatch(this.lookup(needle, haystack), val, haystack)) {
          matches++
        }
      } else {
        if (this.eq(this.lookup(needle, haystack), val))
          matches++
      }
    }
    return matches === matchCount
  }
  static filter(query = {}) {
    let fn = obj => _.filter([obj], _.matches(query)).shift() // lodash filter?
    if (_.isObj(query))
      fn = this.match.bind(this, query)
    else if (_.isFn(query))
      fn = query
    query = this.setOpts(query)
    return this.match.bind(this, query)
  }
  static finish(arr, opts) {
    if (opts.sort) {
      let sort = opts.sort
      let invert = false
      if (sort[0] === '-') {
        sort = sort.slice(1)
        invert = true
      }
      if (invert) {
        arr.sort((A, B) => {
          const a = _.get(A, sort)
          const b = _.get(B, sort)
          return a > b ? -1 : a < b ? 1 : 0
        })
      } else {
        arr.sort((A, B) => {
          const a = _.get(A, sort)
          const b = _.get(B, sort)
          return a > b ? 1 : a < b ? -1 : 0
        })
      }
    }
    if (opts.skip) {
      arr = arr.slice(opts.skip)
    }
    if (opts.select) {
      arr = arr.map(item => _.pick(item, opts.select))
    }
    if (opts.limit) {
      return arr.slice(0, opts.limit)
    }
    return arr
  }
  static find(query, data) {
    const { opts } = this.getOpts(query)
    const fn = this.filter(query)
    data = data.filter(fn)
    return this.finish(data, opts)
  }
  static build(query) {
    return this.Builder.find(query)
  }
}

class QueryBuilder {
  static find(obj = {}) {
    return new this(obj)
  }
  constructor(obj = {}) {
    this.obj = obj
  }
  where(key) {
    const obj = this.obj[key] || {}
    this.obj[key] = obj
    this.key = key
    return this
  }
  equals(val) {
    const key = this.key
    this.obj[this.key].$eq = val
    return this
  }
  gt(num) {
    this.obj[this.key].$gt = num
    return this
  }
  gte(num) {
    this.obj[this.key].$gte = num
    return this
  }
  lt(num) {
    this.obj[this.key].$lt = num
    return this
  }
  lte(num) {
    this.obj[this.key].$lte = num
    return this
  }
  in(arr) {
    this.obj[this.key].$in = arr
    return this
  }
  nin(arr) {
    this.obj[this.key].$nin = arr
    return this
  }
  all(arr) {
    this.obj[this.key].$all = arr
    return this
  }
  limit(num) {
    this.obj.$limit = num
    return this
  }
  skip(num) {
    this.obj.$skip = num
    return this
  }
  sort(key) {
    this.obj.$sort = key
    return this
  }
  select(...keys) {
    this.obj.$select = _.uniq(_.concat(_.toArr(this.obj.$select), keys))
    return this
  }
  filter() {
    return Query.filter(this.obj)
  }
  exec(data) {
    return Query.find(this.obj, data)
  }
}

module.exports = Query
