/**
 * @fileoverview This file contains a collection of utility functions that are 
 * reused throughout the application. These functions include a variety of type 
 * checkers, conversion tools, and other utilities that aid in data manipulation 
 * and validation. This utility library is designed to be environment-agnostic, 
 * providing compatibility with both browser and Node.js environments.
 * 
 * @module Base
 * 
 * @requires lodash
 * @requires bs58
 * @requires lexicographic-integer
 * 
 * @see {@link http://example.com|Optional external link related to this file}
 */

const _ = require('lodash')
const bs58 = require('bs58')
const lexint = require('lexicographic-integer')
const debug = require('./debug')

_._ = _
_.bs58 = bs58
_.lexint = lexint
_.debug = debug
_.BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'

// logging

/**
 * Logs the data.
 * 
 * @type {function}
 * @param {...any} args - Data to log.
 */
const log = _.log = debug
// const log = _.log = console.log.bind(console)

/**
 * Prints the data using the log function.
 * node version uses util to show all
 * 
 * @type {function}
 * @param {...any} args - Data to print.
 */
const print = _.print = _.log

// language

/**
 * Checks if a value is numeric.
 * 
 * @param {any} num - Value to check.
 * @returns {boolean} Whether the value is numeric.
 */
const isNumeric = _.isNumeric = (num) => isFinite(String(num).trim() || NaN)

/**
 * Converts a value to a number.
 * 
 * @param {any} num - Value to convert.
 * @param {number} deci - Number of decimal places.
 * @returns {number} The converted number.
 */
const toNumber = _.toNumber = (num, deci) => 
  parseFloat(parseFloat(String(num).trim()).toFixed(deci))

/**
 * Checks if a value is a valid date.
 * 
 * @param {any} date - Value to check.
 * @returns {boolean} Whether the value is a valid date.
 */
const isDate = _.isDate = (date) => _.isNumeric(new Date(date).getTime())

/**
 * Checks if a string is a valid Base58 string.
 * 
 * @param {string} str - String to check.
 * @returns {boolean} Whether the string is Base58.
 */
const isBs58 = _.isBs58 = str => 
  str.split('').every(char => _.BASE58_ALPHABET.includes(char))

/**
 * Checks if a value is a boolean or "true" / "false" string.
 * 
 * @param {any} bool - Value to check.
 * @returns {boolean} Whether the value is a boolean or "true" / "false" string.
 */
const isBool = _.isBool = (bool) => 
  _.isBoolean(bool) || bool === 'true' || bool === 'false'

/**
 * Converts a value to a boolean.
 * 
 * @param {any} bool - Value to convert.
 * @returns {boolean} The converted boolean.
 */
const toBoolean = _.toBoolean = (bool) =>
  bool === 'true' ? true : bool === 'false' ? false : _.isBool(bool) ? bool : !!bool

/**
 * Checks if a value is an object, excluding arrays, functions, and binary data.
 * 
 * @param {any} input - Value to check.
 * @returns {boolean} Whether the value is an object.
 */ 
const isObj = _.isObj = (input) =>
  _.isObject(input) && !_.isBinary(input) && !_.isArray(input) && !_.isFunction(input)

/**
 * Converts a value to an array. If the value is null or undefined, returns an empty array.
 * 
 * @param {any} obj - Value to convert.
 * @returns {Array} The converted array.
 */
const toArray = _.toArray = (obj) => 
  _.isNil(obj) ? [] : _.isArray(obj) ? obj : [obj]

/**
 * Checks if a value is binary data.
 * 
 * @param {any} value - Value to check.
 * @returns {boolean} Whether the value is binary data.
 */
const isBinary = _.isBinary = (value) => 
  _.isBrowser ? _.isArrayBuffer(value) : _.isBuffer(value)

/**
 * Converts a value to a string. If the value cannot be converted, it returns the original value.
 * 
 * @param {any} obj - Value to convert.
 * @returns {string|any} The converted string or the original value if conversion failed.
 */
const toString = _.toString = (obj) => {
  try {
    obj = obj.toString()
  } catch (err) {}
  return obj
}

/**
 * Transforms any data into a buffer.
 * 
 * @param {any} data - Data to transform into a buffer.
 * @param {string} [encoding] - The encoding to use if not a browser environment.
 * @throws {Error} If data type is unsupported.
 * @returns {Uint8Array|Buffer} The transformed buffer.
 */
const toBuffer = _.toBuffer = (data, encoding) => {
  if (!_.isBrowser) return Buffer.from(data, encoding)
  else if (_.isString(data))
    return new Uint8Array(_.map(data, (char) => char.charCodeAt(0)))
  else if (_.isArray(data)) return new Uint8Array(data)
  throw new Error('Unsupported data type')
}

/**
 * Checks if a value is a truthy boolean.
 * 
 * @param {any} obj - Value to check.
 * @returns {boolean} Whether the value is a truthy boolean.
 */
const truthy = _.truthy = obj => _.isBool(obj) && obj

/**
 * Checks if a value is a falsy boolean.
 * 
 * @param {any} obj - Value to check.
 * @returns {boolean} Whether the value is a falsy boolean.
 */
const falsey = _.falsey = obj => _.isBool(obj) && !obj

/**
 * A no-operation function that does nothing.
 */
const noop = _.noop = () => {}

/**
 * Checks if a value is a function.
 * 
 * @param {any} fn - Value to check.
 * @returns {boolean} Whether the value is a function.
 */
const isFn = _.isFn = fn => (fn && _.isFunction(fn))

/**
 * Checks if a number is odd.
 * 
 * @param {number} num - Number to check.
 * @returns {boolean} Whether the number is odd.
 */
const odd = _.odd = num => !!(num % 2)

/**
 * Checks if a number is even.
 * 
 * @param {number} num - Number to check.
 * @returns {boolean} Whether the number is even.
 */
const even = _.even = num => !_.odd(num)

// environment

/**
 * Detects if the runtime environment is a browser.
 * 
 * @type {boolean}
 */
const isBrowser = _.isBrowser = typeof window !== `undefined`

/**
 * Detects if the runtime environment is Node.js.
 * 
 * @type {boolean}
 */
const isNode = _.isNode = 
  typeof process !== 'undefined' 
  && process.versions != null 
  && process.versions.node != null

/**
 * Detects if the device is a mobile device.
 * 
 * @type {boolean}
 */
const isMobile = _.isMobile = (() => {
  if (!_.isBrowser)
    return false
  return (/Android|webOS|iPhone|iPad|iPod|BlackBerry/i.test(navigator.userAgent) ||
         (/Android|webOS|iPhone|iPad|iPod|BlackBerry/i.test(navigator.platform)))
})()

/**
 * Sets the URL object based on the runtime environment.
 * 
 * @type {URL}
 */
const URL = _.URL = _.isBrowser ? window.URL : require('url').URL

/**
 * Checks if a string is a valid URL.
 * 
 * @param {string} url - String to check.
 * @returns {boolean} Whether the string is a valid URL.
 */
const isUrl = _.isUrl = (url) => {
  try {
    new _.URL(url)
    return true
  } catch (err) {
    return false
  }
}

// strings

/**
 * Capitalizes the first character of a string.
 * 
 * @param {string} str - String to capitalize.
 * @returns {string} The capitalized string.
 */
const ucf = _.ucf = str => 
  str.toString()[0].toUpperCase() + str.toString().slice(1)

/**
 * Converts a string to upper case.
 * 
 * @param {string} str - String to convert.
 * @returns {string} The converted string.
 */
const uc = _.uc = str => str.toUpperCase()

/**
 * Converts a string to lower case.
 * 
 * @param {string} str - String to convert.
 * @returns {string} The converted string.
 */
const lc = _.lc = str => str.toLowerCase()

/**
 * Pads a string from the start with spaces until it reaches the specified length.
 * 
 * @param {string} str - String to pad.
 * @param {number} len - Target length.
 * @returns {string} The padded string.
 */
const pad = _.pad = (str, len) => str.padStart(len)

/**
 * Removes leading white space from a string.
 * 
 * @param {string} str - String to trim.
 * @returns {string} The trimmed string.
 */
const unpad = _.unpad = str => str.trimStart()

// utility

/**
 * Resolves a promise after a delay.
 * 
 * @param {number|string} [ms=0] - Delay in milliseconds or a string representing time.
 * @returns {Promise<void>} A promise that resolves after the specified delay.
 */
const sleep = _.sleep = (ms = 0) => 
  new Promise(resolve => setTimeout(resolve, _.isString(ms) ? _.str2ms(ms) : ms))

/**
 * Defines a new property directly on an object, or modifies an existing property 
 * on an object, and returns the object.
 * 
 * @param {Object} obj - The object on which to define or modify the property.
 * @param {string|Symbol} key - The name or Symbol of the property to be defined or modified.
 * @param {any} value - The value to set the property to.
 * @param {Object} [opts={}] - The optional attributes for the property descriptor.
 * @returns {Object} The object on which the property was defined.
 */
const objProp = _.objProp = (obj, key, value, opts = {}) => {
  if (!_.isNil(opts.show)) {
    opts.enumerable = opts.show
    delete opts.show
  }
  if (!_.isNil(opts.enum)) {
    opts.enumerable = opts.enum
    delete opts.enum
  }
  if (value !== undefined) opts.value = value
  Object.defineProperty(obj, key, opts)
}

/**
 * Creates a Proxy for an object to customize the behavior of fundamental operations (e.g., property lookup, assignment, enumeration, function invocation, etc).
 * 
 * @param {Object} obj - The target object for the Proxy.
 * @param {function} callback - The handler callback for the get trap.
 * @returns {Proxy} The Proxy for the object.
 */
const proxyGet = _.proxyGet = (obj, callback) => new Proxy(obj, {
  get: (target, prop, receiver) => {
    if (Reflect.has(target, prop) || !_.isString(prop) || prop === 'then')
      return Reflect.get(target, prop, receiver)
    return callback(target, prop, receiver)
  },
})

/**
 * Creates a Proxy for an object to customize its instantiation behavior.
 * 
 * @param {Object} obj - The target object for the Proxy.
 * @returns {Proxy} The Proxy for the object.
 */
const proxyNew = _.proxyNew = (obj) => new Proxy(obj, {
  apply: (target, __, args) => new target(...args),
})

/**
 * Returns the names of all methods of an object.
 * 
 * @param {Object} obj - The object to get methods from.
 * @param {boolean|string[]} [ignore] - Whether to ignore built-in methods or an array of method names to ignore.
 * @param {boolean} [underscore=false] - Whether to include methods that start with an underscore.
 * @param {boolean} [getters=false] - Whether to include getter methods.
 * @returns {string[]} The method names.
 */
const getMethods = _.getMethods = (obj, ignore, underscore = false, getters = false) => {
  const orig = obj
  let keys = []
  let include = _.isBoolean(ignore) && ignore
  do {
    keys = keys.concat(Object.getOwnPropertyNames(obj))
    if (!getters)
      keys = keys.filter(k => !(Object.getOwnPropertyDescriptor(obj, k) || {}).get)
    if (!underscore)
      keys = keys.filter(k => !k.startsWith('_'))
    if (ignore && !_.isBoolean(ignore))
      keys = keys.filter(k => !_.toArray(ignore).includes(k))
    obj = Object.getPrototypeOf(obj)
  } while (obj && obj.constructor && obj.constructor.name !== 'Object')
  return Array.from(new Set(keys))
    .filter(k => typeof orig[k] === 'function')
    .filter(k => include ? true : k !== 'constructor')
}

/**
 * Wraps a value in a promise if it is not already a promise, allowing it to be used as a promise.
 * 
 * @param {any} res - The value or promise.
 * @returns {Promise<any>} The promise.
 */
const respond = _.respond = (res) => new Promise((resolve, reject) => {
  return res instanceof Promise ? res.then(resolve).catch(reject) : resolve(res)
})

// encoding

/**
 * Encodes a buffer to a Base58 or hex string.
 * 
 * @param {Buffer} buff - Buffer to encode.
 * @param {boolean} [hex] - Whether to encode to hex instead of Base58.
 * @returns {string} The encoded string.
 */
const encode = _.encode = (buff, hex) => {
  if (hex)
    return buff.toString('hex')
  return bs58.encode(buff)
}

/**
 * Decodes a Base58 or hex string into a buffer.
 * 
 * @param {string} str - String to decode.
 * @param {boolean|string} [toBuffer=true] - Whether to convert the result to a buffer. If 'hex', the result is converted to a hex buffer.
 * @throws {Error} If the input string is not a string.
 * @returns {Buffer} The decoded buffer.
 */
const decode = _.decode = (str, toBuffer = true) => {
  if (!_.isString(str))
    throw new Error(`cannot decode a non-string`)
  if (toBuffer === 'hex')
    return _.toBuffer(str, 'hex')
  else if (toBuffer)
    return _.toBuffer(bs58.decode(str))
  return bs58.decode(str)
}

/**
 * Converts a string representing time to milliseconds.
 * 
 * @param {string} str - String to convert.
 * @param {Object} [obj={}] - Object to populate with conversion details.
 * @returns {number} The converted milliseconds.
 */
const str2ms = _.str2ms = (str, obj = {}) => {
  if (_.isNumeric(str))
    if (_.isInteger(_.toNumber(str)))
      return _.toNumber(str)
    else
      return str
  else if (!_.isString(str))
    return str
  let unit
  if (str.match(/ms$/))
    unit = str.slice(-2)
  else
    unit = str.slice(-1)
  const val = _.toNumber(str.replace(unit, ''))
  if (!_.isNumeric(val))
    return str

  const ms = {
    ms: 0.001,
    s: 1,
    m: 60,
    h: 60 * 60,
    d: 60 * 60 * 24,
    w: 60 * 60 * 24 * 7,
    M: 60 * 60 * 24 * 30
  }[unit] * val * 1000

  const [ singular, plural ] = {
    ms: [ 'milisecond', 'miliseconds' ],
    s: [ 'second', 'seconds' ],
    m: [ 'minute', 'minutes' ],
    h: [ 'hour', 'hours' ],
    d: [ 'day', 'days' ],
    w: [ 'week', 'weeks' ],
    M: [ 'month', 'months' ],
  }[unit]

  Object.assign(obj, { ms, unit, val, singular, plural })

  return ms
}

/**
 * Sorts the characters of a string or a JSON stringified object.
 * 
 * @param {string|Object} data - String or object to sort characters of.
 * @returns {string} The string with sorted characters.
 */
const sortChars = _.sortChars = (data) => {
  if (!_.isString(data))
    data = JSON.stringify(data)
  return data.split('').sort().join('')
}

// execution

/**
 * Executes a function and logs its execution time.
 * 
 * @param {function} exec - Function to execute.
 * @returns {Promise<any>} A promise that resolves with the result of the function.
 */
_.executor = async exec => {
  const side = _.isBrowser ? 'browser' : 'server'
  const begin = Date.now()
  _.log(`\n<<< tonight is your night bro! [${side}] >>>\n`)
  return Promise.resolve(exec())
    .then(() => _.log(`\n>>> execution time: ${Date.now() - begin}ms <<<\n`))
    .catch(console.error)
}

module.exports = _

