/**
 * @fileoverview Provides an event emitter mixin class for applications.
 *
 * This module exports a mixin class `Evented` for creating objects with 
 * capabilities of event-driven behavior. It is part of the `basd` project 
 * and heavily relies on other foundational parts of the project.
 * 
 * The `Evented` class and its mixin function are exported from the module.
 * @module Evented
 */

const { _, log } = require('./base')
const EVENTS = Symbol('events')

/**
 * @function EventedMixin
 * @class Evented
 * @param {Class} Class - The class to extend with event capabilities.
 * @returns {Evented} - A class extended with event capabilities.
 */
const EventedMixin = (Class = class {}) => 
class Evented extends Class {
  /**
   * Constructs an Evented instance.
   * @param {...*} args - The arguments to pass to parent class.
   */
  constructor(...args) {
    super(...args)
    Object.defineProperty(this, EVENTS, { value: new Map, enumerable: false })
  }

  /**
   * @method on
   * @description Registers a handler function for a specific topic.
   * @param {string} topic - The event topic.
   * @param {Function} handler - The handler function for the event.
   * @returns {Function} - The off function, pre-bound with topic and handler.
   */
  on(topic, handler) {
    if (!this[EVENTS].has(topic))
      this[EVENTS].set(topic, new Set)
    if (typeof handler === 'undefined')
      return this[EVENTS].get(topic)
    this[EVENTS].get(topic).add(handler)
    return this.off.bind(this, topic, handler)
  }

  /**
   * @method onDefault
   * @description Registers a default handler function for a specific topic.
   * @param {string} topic - The event topic.
   * @param {Function} handler - The handler function for the event.
   */
  onDefault(topic, handler) {
    this.on(`__default__${topic}`, handler)
  }

  /**
   * @method off
   * @description Deregisters a handler function from a specific topic.
   * @param {string} topic - The event topic.
   * @param {Function} handler - The handler function to remove.
   */
  off(topic, handler) {
    if (!this[EVENTS].has(topic))
      return
    const set = this[EVENTS].get(topic)
    if (typeof handler === 'undefined')
      return this[EVENTS].delete(topic)
    if (set.has(handler))
      set.delete(handler)
  }

  /**
   * @method once
   * @description Registers a handler function for a specific topic for once.
   * @param {string} topic - The event topic.
   * @param {Function} handler - The handler function for the event.
   */
  once(topic, handler) {
    const _handler = (...args) => {
      handler(...args)
      this.off(topic, _handler)
    }
    this.on(topic, _handler)
  }

  /**
   * @method emit
   * @description Emits an event on a specific topic, executing all handlers.
   * @param {string} topic - The event topic.
   * @param {...*} args - The arguments to pass to the handlers.
   */
  emit(topic, ...args) {
    if (!this[EVENTS].has(topic))
      return !topic.startsWith('__default__') ? this.emit(`__default__${topic}`, ...args) : null
    const set = this[EVENTS].get(topic)
    try {
      if (set.size === 1)
        for (const fn of set)
          return fn(...args)
      for (const fn of set)
        fn(...args)
    } catch (error) {
      this.emit('error', error)
    }
  }

  /**
   * @method has
   * @description Checks if there are any handlers for a specific topic.
   * @param {string} topic - The event topic.
   * @returns {boolean} - Whether any handlers exist for the given topic.
   */
  has(topic) {
    try {
      return this.size(topic) > 0
    } catch (error) {}
    return false
  }

  /**
   * @method size
   * @description Returns the number of handlers for a specific topic.
   * @param {string} topic - The event topic.
   * @returns {number} - The number of handlers for the given topic.
   */
  size(topic) {
    if (typeof topic === 'undefined')
      return this[EVENTS].size
    if (!this[EVENTS].has(topic))
      throw new Error(`topic "${topic}" not found`)
    return this[EVENTS].get(topic).size
  }

  /**
   * @method clear
   * @description Clears all handlers for a specific topic.
   * @param {string} topic - The event topic.
   */
  clear(topic) {
    if (topic)
      return this.off(topic)
    for ([ topic ] of this[EVENTS])
      this.off(topic)
  }
}

const Evented = EventedMixin()
Object.defineProperty(Evented, 'mixin', { value: Class => EventedMixin(Class) })
module.exports = Evented

// module.exports = new Proxy(class {
//   static get mixin() { return EventedMixin }
//   static get Class() { return EventedMixin() }
//   }, {
//   construct: (Class, args, target) => _.nameExtend(target.name, EventedMixin(), ...args),
//   apply: (Class, target, args) => Reflect.construct(EventedMixin(), args)
// })
