/**
 * @fileoverview This file defines the Heap and QueueItem classes, used to
 * implement a heap data structure for use within the @basd/base submodule.
 * This heap implementation is a key part of the @basd foundational library.
 * @module Heap
 */

const { _, log } = require('./base')

const items = Symbol('items')
const comparator = Symbol('comparator')

/**
 * Represents an item in a heap.
 */
class QueueItem {
  /**
   * Constructs a QueueItem instance.
   * @param {any} value The value of the item.
   * @param {number} key The sorting key of the item.
   */
  constructor(value, key = 1) {
    this.key = key
    this.value = value
  }

  /**
   * Returns the key of the item.
   * @return {number} The key of the item.
   */
  getKey() {
    return this.key
  }

  /**
   * Returns the value of the item.
   * @return {any} The value of the item.
   */
  getValue() {
    return this.value
  }

  /**
   * Creates a new QueueItem instance with the same key and value as this one.
   * @return {QueueItem} The cloned QueueItem.
   */
  clone() {
    return new QueueItem(this.key, this.value)
  }
}

/**
 * Represents a heap data structure.
 */
class Heap {
  /**
   * Constructs a Heap instance.
   * @param {object} opts The options for the heap, including initial items and
   * a custom comparator function.
   */
  constructor(opts = {}) {
    _.objProp(this, items, opts.items || [], { writable: true, enumerable: true })
    _.objProp(this, comparator, opts.comparator
      ? opts.comparator.bind(this)
      : this._defaultComparator.bind(this))
    if (typeof opts.items === 'object') {
      this.heapify(opts.items)
    }
  }

  /**
   * Default comparator function that compares items based on their keys.
   * @private
   * @param {QueueItem} a The first item.
   * @param {QueueItem} b The second item.
   * @return {number} A number indicating the sort order.
   */
  _defaultComparator(a, b) {
    return a.getKey() - b.getKey()
  }

  /**
   * Inserts an item into the heap based on the result of the comparator function.
   * @private
   * @param {QueueItem} item The item to be inserted.
   */
  _comparator(item) {
    let index = this[items].findIndex((existingItem) => this[comparator](existingItem, item) > 0)
    if (index >= 0) {
      this[items].splice(index, 0, item)
    } else {
      this[items].push(item)
    }
  }

  /**
   * Creates a valid heap from an unordered array of items in linear time.
   * @param {Array.<{key: number, value: any}>} unorderedItems An array of objects,
   * each with a 'key' for ordering and a 'value'.
   */
  heapify(unorderedItems) {
    this[items] = unorderedItems.map(item => new QueueItem(item.value, item.key))
    for (let ii = Math.floor(this.length / 2) - 1; ii >= 0; ii--) {
      this._siftDown(ii)
    }
  }

  /**
   * Moves the item at the provided index down to its proper place in the heap.
   * @private
   * @param {number} index The index of the item to be moved.
   */
  _siftDown(index) {
    let childIndex = 2 * index + 1
    if (childIndex >= this.length) {
      return
    }
    // If the node has two children and the second child is smaller than the first,
    // point to the second child
    if (childIndex + 1 < this.length
      && this[comparator](this[items][childIndex], this[items][childIndex + 1]) > 0) {
      childIndex++
    }

    // If the node is smaller than its smallest child, swap them
    if (this[comparator](this[items][childIndex], this[items][index]) < 0) {
      [this[items][childIndex], this[items][index]] = [this[items][index], this[items][childIndex]]
      this._siftDown(childIndex)
    }
  }

  /**
   * Inserts an item into the heap.
   * @param {any} value The value of the item.
   * @param {number} key The key of the item.
   */
  insert(value, key) {
    if (value === undefined || value === null) {
      return
    }
    this._comparator(new QueueItem(value, key))
  }

  /**
   * Removes and returns the item at the top of the heap.
   * @return {any|null} The value of the removed item or null if the heap is empty.
   */
  remove() {
    if (this.isEmpty()) {
      return null
    }
    const item = this[items].shift()
    if (!item) {
      return null
    }
    return item.getValue()
  }

  /**
   * Returns the value of the item at the top of the heap without removing it.
   * @return {any|null} The value of the top item or null if the heap is empty.
   */
  peek() {
    if (this.isEmpty()) {
      return null
    }
    return this[items][0].getValue()
  }

  /**
   * Returns the key of the item at the top of the heap.
   * @return {number|undefined} The key of the top item or undefined if the heap is empty.
   */
  peekKey() {
    return this[items][0] && this[items][0].getKey()
  }

  /**
   * Checks if the heap is empty.
   * @return {boolean} True if the heap is empty, false otherwise.
   */
  isEmpty() {
    return this.length === 0
  }

  /**
   * Clears all items from the heap.
   */
  clear() {
    this[items].length = 0
  }

  /**
   * Returns the number of items in the heap.
   * @return {number} The number of items in the heap.
   */
  get length() {
    return this[items].length
  }
}

module.exports = Heap
