/**
 * @fileoverview This module, Temporal, is an event simulator for network activity. 
 * It can simulate different network events such as intervals, timeouts, and immediate
 * executions, that are crucial in various networking scenarios.
 * 
 * @module Temporal
 */

const { _, log } = require('./base')
const Heap = require('./heap')

/**
 * @class
 * @classdesc The Temporal class simulates the passage of time in a network environment.
 * It is equipped to handle different types of events such as intervals, timeouts, and 
 * immediate executions.
 * 
 * @property {number} began - The timestamp of when the Temporal instance started.
 * @property {number} time - The current time in the Temporal instance.
 * @property {number} maxrun - The maximum duration for the Temporal instance to run.
 * @property {number} speed - The speed at which time passes in the Temporal instance.
 * @property {boolean} running - Whether the Temporal instance is currently running.
 * @property {Heap} _heap - A heap data structure to hold the events in a priority queue.
 * @property {number} _timer - A reference to the setTimeout that's currently running.
 */
class Temporal {
  /**
   * @constructor
   * @param {Object} opts - The options to configure the Temporal instance.
   */
  constructor(opts = {}) {
    this.began = opts.began || Date.now()
    this.time = opts.time || 0
    this.maxrun = opts.maxrun || 0
    this.speed = !_.isNil(opts.speed) ? opts.speed : 1
    _.objProp(this, 'running', false, { writable: true })
    _.objProp(this, '_heap', new Heap, { enumerable: false })
    _.objProp(this, '_timer', null, { writable: true })
  }

  /**
   * Returns the current time in the Temporal instance.
   * 
   * @returns {number} The current time in the Temporal instance.
   */
  now() {
    return this.began + this.time
  }

  /**
   * Starts the Temporal instance.
   *
   * @param {number} speed - The speed at which time should pass.
   * @param {string|number} ms - The duration the Temporal instance should run.
   * @returns {Promise} A promise that resolves when the Temporal instance starts.
   */
  async start(speed, ms) {
    this.running = true
    ms = _.str2ms(ms || Infinity)
    this.speed = !_.isNil(speed) ? speed : this.speed
    this.maxrun = this.time + ms
    return this.run(ms)
  }

  /**
   * Runs the Temporal instance for a given duration.
   *
   * @param {string|number} ms - The duration the Temporal instance should run.
   * @returns {Promise} A promise that resolves when the Temporal instance starts running.
   */
  run(ms) {
    const next = this._next(Math.min(this.time + ms, this.maxrun))
    if (!next) {
      this.running = false
      return
    }
    const speed = this.speed
    const { time, event } = next
    const { type, delay, resolve } = event
    let duration = 0
    let begin
    if (speed) {
      const prev = this.time
      duration = time - prev
      begin = Date.now()
    }
    this.time = next.time
    const prom = _.respond(next.event.run(this))
    if (speed) {
      const finish = Date.now()
      const delta = finish - begin
      let remain = duration - delta
      remain = remain >= 0 ? remain : 0
      duration = speed === 0 ? 0 : Math.floor(remain / speed)
    }
    return new Promise(done => {
      this._timer = setTimeout(() => {
        resolve(prom)
        done(this.run(ms))
      }, duration)
    })
  }

  /**
   * Enqueues a new event in the Temporal instance.
   *
   * @private
   * @param {Object} event - The event to enqueue.
   */
  _enqueue(event) {
    this._heap.insert(event, event.delay + this.time)
  }

  /**
   * Wraps the provided function and binds the given context and arguments.
   *
   * @private
   * @param {Function} fn - The function to wrap.
   * @param {Object} ctx - The context to bind to the function.
   * @param {...*} args - The arguments to bind to the function.
   * @returns {Function} The wrapped function.
   */
  _getFn(fn, ctx, ...args) {
    if (_.isObj(ctx))
      fn = fn.bind(ctx, ...args)
    return fn
  }

  /**
   * Returns the next event that should be executed in the Temporal instance.
   *
   * @private
   * @param {number} maxtime - The maximum time an event can have to be considered.
   * @returns {Object|false} The next event to be executed, or false if no suitable event was found.
   */
  _next(maxtime) {
    const time = this._heap.peekKey()
    if (_.isNil(time) || time > maxtime)
      return false
    return { time, event: this._heap.remove() }
  }

  /**
   * Executes a function immediately in the Temporal instance.
   *
   * @param {Function} fn - The function to execute.
   * @param {Object} ctx - The context to bind to the function.
   * @param {...*} args - The arguments to pass to the function.
   * @returns {Promise} A promise that resolves when the function has been executed.
   */
  exec(fn, ctx, ...args) {
    fn = this._getFn(fn, ctx, ...args)
    return new Promise(resolve => {
      this._enqueue({
        type: 'immediate',
        delay: 0,
        run: timer => fn(),
        resolve,
      })
    })
  }

  /**
   * Executes a function after a specified delay in the Temporal instance.
   *
   * @param {string|number} delay - The delay after which the function should be executed.
   * @param {Function} fn - The function to execute.
   * @param {Object} ctx - The context to bind to the function.
   * @param {...*} args - The arguments to pass to the function.
   * @returns {Promise} A promise that resolves when the function has been executed.
   */
  delay(delay, fn, ctx, ...args) {
    fn = this._getFn(fn, ctx, ...args)
    return new Promise(resolve => {
      this._enqueue({
        type: 'timeout',
        delay: _.str2ms(delay),
        run: timer => fn(),
        resolve,
      })
    })
  }

  /**
   * Executes a function repeatedly, with a specified delay between executions, in the Temporal instance.
   *
   * @param {string|number} delay - The delay between executions of the function.
   * @param {Function} fn - The function to execute.
   * @param {Object} ctx - The context to bind to the function.
   * @param {...*} args - The arguments to pass to the function.
   * @returns {Promise} A promise that resolves when the function has started being executed repeatedly.
   */
  tick(delay, fn, ctx, ...args) {
    fn = this._getFn(fn, ctx, ...args)
    return new Promise(resolve => {
      this._enqueue({
        type: 'interval',
        delay: _.str2ms(delay),
        run (timer) {
          return _.respond(fn()).then(delay => {
            if (_.isInteger(delay))
              this.delay = delay
            timer._enqueue(this)
          })
        },
        resolve,
      })
    })
  }

  /**
   * Stops the Temporal instance.
   */
  stop() {
    clearTimeout(this._timer)
    this.maxrun = this.time
    this.running = false
  }
}

module.exports = Temporal
