/**
 * @fileoverview Defines Timer and Timely classes to create and manage 
 * timer events with a variety of options.
 * @module Timely
 * @requires ./base
 */

const { _, log } = require('./base')

/**
 * The Timer class wraps around timer functions and provides utility 
 * methods for time calculations.
 */
class Timer {
  /**
   * Constructs a new Timer object.
   * @param {Function} timer - Function to set the timer.
   * @param {Function} clear - Function to clear the timer.
   * @param {number} duration - Timer duration in milliseconds.
   * @param {Function} fn - Function to be executed when timer finishes.
   */
  constructor(timer, clear, duration, fn) {
    this.start = Date.now()
    this.duration = duration
    this.clear = clear
    _.objProp(this, 'timer', timer, { writable: true })
    this.fns = [fn]
  }
  /**
   * Calculate the remaining time on the timer.
   * @return {number} Remaining time in milliseconds.
   */
  remaining() {
    return this.duration - this.taken()
  }
  /**
   * Calculate the elapsed time since the timer started.
   * @return {number} Elapsed time in milliseconds.
   */
  taken() {
    return Date.now() - this.start
  }
}

/**
 * The Timely class provides methods for managing timers including
 * setTimeout, setInterval, and setImmediate.
 */
class Timely {
  static Timely = Timely
  static Timer = Timer
  /**
   * Constructs a new Timely object.
   * @param {Object} [opts={}] - An options object that includes custom 
   * timer functions.
   * @param {*} [ctx] - The context for 'this'.
   */
  constructor(opts = {}, ctx) {
    this.timers = {}
    _.objProp(this, 'ctx', ctx || this, { writable: true })
    const timeout = opts.setTimeout || setTimeout
    const _setTimeout = (fn, time) => timeout(fn, time)
    _.objProp(this, '_setTimeout', _setTimeout)
    // _.objProp(this, '_setTimeout', opts.setTimeout || setTimeout)
    _.objProp(this, '_clearTimeout', opts.clearTimeout || clearTimeout)
    _.objProp(this, '_setImmediate', opts.setImmediate || setImmediate)
    _.objProp(this, '_clearImmediate', opts.clearImmediate || clearImmediate)
    _.objProp(this, '_setInterval', opts.setInterval || setInterval)
    _.objProp(this, '_clearInterval', opts.clearInterval || clearInterval)
  }

  /**
   * Retrieves a named function, executes it, and either clears or resets the timer.
   * @param {string} name - Name of the timer function.
   * @param {boolean} clear - Whether to clear or reset the timer.
   * @return {Function} Function that will execute the desired action when called.
   */
  getFn(name, clear) {
    return () => {
      if (!this.timers[name])
        return
      const timer = this.timers[name]
      const fns = timer.fns.slice()
      if (clear)
        this.clear(name)
      else
        timer.start = Date.now()
      for (let ii = 0; ii < fns.length; ii++) {
        fns[ii].call(this.ctx)
      }
    }
  }

  /**
   * Set a named timeout function.
   * @param {string} name - Name of the timer function.
   * @param {Function} fn - Function to execute after the timeout.
   * @param {number} time - Time in milliseconds after which to execute the function.
   * @return {this} Current instance.
   */
  setTimeout(name, fn, time) {
    if (this.timers[name]) {
      this.timers[name].fns.push(fn)
      return this
    }
    const duration = _.str2ms(time)
    this.timers[name] = new Timer(
      this._setTimeout(this.getFn(name, true), duration),
      this._clearTimeout,
      duration,
      fn
    )
    return this
  }

  /**
   * Set a named interval function.
   * @param {string} name - Name of the timer function.
   * @param {Function} fn - Function to execute on each interval.
   * @param {number} time - Time in milliseconds for the interval.
   * @return {this} Current instance.
   */
  setInterval(name, fn, time) {
    if (this.timers[name]) {
      this.timers[name].fns.push(fn)
      return this
    }
    const duration = _.str2ms(time)
    this.timers[name] = new Timer(
      this._setInterval(this.getFn(name), duration),
      this._clearInterval,
      duration,
      fn
    )
    return this
  }

  /**
   * Set a named immediate function.
   * @param {string} name - Name of the immediate function.
   * @param {Function} fn - Function to execute immediately.
   * @return {this} Current instance.
   */
  setImmediate(name, fn) {
    if ('function' !== typeof setImmediate)
      return this.setTimeout(name, fn, 0)
    if (this.timers[name]) {
      this.timers[name].fns.push(fn)
      return this
    }
    this.timers[name] = new Timer(
      this._setInterval(this.getFn(name, true)),
      this._clearImmediate,
      0,
      fn
    )
    return this
  }

  /**
   * Check if a timer is active.
   * @param {string} name - Name of the timer.
   * @return {boolean} Whether the timer is active.
   */
  active(name) {
    return name in this.timers
  }

  /**
   * Clear one or more timers.
   * @param {...string} args - Names of the timers to clear.
   * @return {this} Current instance.
   */
  clear(...args) {
    let timer
    if (args.length === 1 && _.isString(args[0]))
      args = args[0].split(/[, ]+/)
    if (!args.length)
      args = _.keys(this.timers)
    for (let ii = 0; ii < args.length; ii++) {
      timer = this.timers[args[ii]]
      if (!timer)
        continue
      timer.clear(timer.timer)
      timer.fns = timer.timer = timer.clear = null
      delete this.timers[args[ii]]
    }
    return this
  }

  /**
   * Adjust the duration of a named timer.
   * @param {string} name - Name of the timer.
   * @param {number} time - New duration for the timer in milliseconds.
   * @return {this} Current instance.
   */
  adjust(name, time) {
    const duration = _.str2ms(time)
    const timer = this.timers[name]
    if (!timer)
      return this
    const interval = timer.clear === this._clearInterval
    timer.clear(timer.timer)
    timer.start = Date.now()
    timer.duration = duration
    const fn = this.getFn(name, !interval)
    const exec = interval ? this._setInterval : this._setTimeout
    timer.timer = exec(fn, duration)
    // log({ fn, setter, duration })
    // timer.timer = (interval ? this._setInterval : this._setTimeout)(this.getFn(name, !interval), duration)

    return this
  }

  /**
   * End all timers and clear the context.
   * @return {boolean} Whether the context was cleared.
   */
  end() {
    if (!this.ctx)
      return false
    this.clear()
    this.ctx = null
    this.timers = {}
    return true
  }
}

module.exports = Timely
