/**
 * @fileoverview Timestamp - A module that generates universally unique hex string 
 * timestamps. The timestamps generated are unique across instances and even on 
 * the same millisecond. This module is part of the `basd` library, which is a 
 * foundational lib for many other projects.
 * @module Timestamp
 */

const { _, log } = require('./base')

/**
 * A class representing a timestamp generator. The generated timestamps consist of 
 * a hexadecimal representation of the current time in milliseconds and an instance 
 * counter.
 */
class Timestamp {
  /**
   * Create a timestamp generator.
   *
   * @param {string} separator - The separator used in the timestamp. Default is '_'.
   * @param {string} prefix - The prefix added before the timestamp. Default is empty.
   */
  constructor(separator = '_', prefix = '') {
    if (this.constructor.instance)
      return this.constructor.instance
    this.separator = separator
    this.prefix = prefix
    this.lastTimestamp = 0
    this.counter = 0
    _.objProp(this.constructor, 'instance', this, { writable: true })
  }

  /**
   * Generate a new timestamp.
   *
   * @returns {string} The generated timestamp.
   */
  now() {
    const now = Date.now()
    if (now === this.lastTimestamp) {
      this.counter++
    } else {
      this.counter = 0
      this.lastTimestamp = now
    }
    const counterPadded = _.lexint.pack(this.counter, 'hex')
    return `${this.prefix}${_.lexint.pack(now, 'hex')}${this.separator}${counterPadded}`
  }

  /**
   * Static method to generate a new timestamp.
   *
   * @param {object} opts - Options for creating the timestamp.
   * @returns {string} The generated timestamp.
   */
  static now(opts) {
    const obj = new this(opts)
    return obj.now()
  }

  /**
   * Static method to create a bound version of the now function.
   *
   * @param {object} opts - Options for creating the timestamp.
   * @returns {function} Bound version of the now function.
   */
  static create(opts) {
    const obj = new this(opts)
    return obj.now.bind(obj)
  }

  /**
   * Static method to create a timestamp with ':' as the separator.
   *
   * @returns {function} Bound version of the now function with ':' as the separator.
   */
  static get mlts() {
    return this.create(':')
  }

  /**
   * Parse a timestamp back into the original time in milliseconds.
   *
   * @param {string} ts - The timestamp to be parsed.
   * @returns {number} The original time in milliseconds.
   */
  parse(ts) {
    return this.constructor.parse(ts, this.separator, this.prefix)
  }

  /**
   * Static method to parse a timestamp back into the original time in milliseconds.
   *
   * @param {string} ts - The timestamp to be parsed.
   * @param {string} separator - The separator used in the timestamp. Default is '_'.
   * @param {string} prefix - The prefix added before the timestamp.
   * @returns {number} The original time in milliseconds.
   */
  static parse(ts, separator = '_', prefix) {
    if (prefix)
      ts = ts.substring(prefix.length)
    const [packedTimestamp] = ts.split(separator)
    return _.lexint.unpack(packedTimestamp)
  }
}

module.exports = Timestamp
