const _ = require('../lib/base')

describe('@basd/base', () => {

  describe('isNumeric', () => {
    it('should return true for numeric values', () => {
      expect(_.isNumeric(123)).to.be.true
      expect(_.isNumeric('123')).to.be.true
    })

    it('should return false for non-numeric values', () => {
      expect(_.isNumeric('abc')).to.be.false
      expect(_.isNumeric(null)).to.be.false
    })
  })

  describe('toNumber', () => {
    it('should convert string to number', () => {
      expect(_.toNumber('123', 2)).to.equal(123.00)
    })

    it('should return NaN for non-numeric string', () => {
      expect(_.toNumber('abc', 2)).to.be.NaN
    })
  })

  describe('isDate', () => {
    it('should return true for valid date', () => {
      expect(_.isDate(new Date())).to.be.true
      expect(_.isDate("2022-12-25")).to.be.true
    })

    it('should return false for invalid date', () => {
      expect(_.isDate('2020-13-01')).to.be.false
      expect(_.isDate("not a date")).to.be.false
      expect(_.isDate(undefined)).to.be.false
    })
  })

  describe('isBs58', () => {
    it('should return true for base58 encoded string', () => {
      const bs58Str = _.bs58.encode(Buffer.from('hello world'))
      expect(_.isBs58(bs58Str)).to.be.true
    })

    it('should return false for non-base58 encoded string', () => {
      expect(_.isBs58('hello world')).to.be.false
    })
  })

  describe('isBool', () => {
    it('should return true for boolean input', () => {
      expect(_.isBool(true)).to.be.true
      expect(_.isBool(false)).to.be.true
      expect(_.isBool('true')).to.be.true
      expect(_.isBool('false')).to.be.true
    })

    it('should return false for non-boolean input', () => {
      expect(_.isBool('not a boolean')).to.be.false
      expect(_.isBool(undefined)).to.be.false
      expect(_.isBool(null)).to.be.false
    })
  })

  describe('toBoolean', () => {
    it('should convert string to boolean', () => {
      expect(_.toBoolean('true')).to.be.true
      expect(_.toBoolean('false')).to.be.false
    })

    it('should return boolean for boolean values', () => {
      expect(_.toBoolean(true)).to.be.true
      expect(_.toBoolean(false)).to.be.false
    })
  })

  describe('isObj', () => {
    it('should return true for object', () => {
      expect(_.isObj({})).to.be.true
    })

    it('should return false for non-object', () => {
      expect(_.isObj('string')).to.be.false
      expect(_.isObj(null)).to.be.false
    })
  })

  describe('toArray', () => {
    it('should convert non-array to array', () => {
      expect(_.toArray('string')).to.eql(['string'])
      expect(_.toArray(null)).to.eql([])
    })

    it('should return array for array', () => {
      expect(_.toArray(['string'])).to.eql(['string'])
    })
  })

  describe('isBinary', () => {
    it('should return true for binary values', () => {
      if (!_.isBrowser) expect(_.isBinary(Buffer.from('test'))).to.be.true
    })

    it('should return false for non-binary values', () => {
      expect(_.isBinary('string')).to.be.false
    })

    it('should return true if the input is a Buffer and running in Node.js', function() {
      if (_.isNode) {
        const buffer = Buffer.from('binary data')
        const result = _.isBinary(buffer)
        expect(result).to.be.true
      }
    })

    it('should return true if the input is an ArrayBuffer and running in the browser', function() {
      if (_.isBrowser) {
        const buffer = new ArrayBuffer(8)
        const result = _.isBinary(buffer)
        expect(result).to.be.true
      }
    })
  })

  describe('toString', () => {
    it('should convert object to string', () => {
      expect(_.toString({ a: 1 })).to.equal('[object Object]')
      expect(_.toString(123)).to.equal('123')
    })
  })

  describe('toBuffer', () => {
    it('should convert string to buffer', () => {
      if (!_.isBrowser) expect(_.toBuffer('string')).to.deep.equal(Buffer.from('string'))
    })

    it('should convert array to buffer', () => {
      if (_.isBrowser) expect(_.toBuffer([0, 1, 2, 3])).to.deep.equal(new Uint8Array([0, 1, 2, 3]))
    })

    it('should throw error for unsupported data type', () => {
      expect(() => _.toBuffer({})).to.throw()
    })
  })

  describe('isFn', () => {
    it('should return true for function', () => {
      expect(_.isFn(() => {})).to.be.true
    })

    it('should return false for non-function', () => {
      expect(_.isFn('string')).to.be.false
    })
  })

  describe('odd', () => {
    it('should return true for odd number', () => {
      expect(_.odd(1)).to.be.true
    })

    it('should return false for even number', () => {
      expect(_.odd(2)).to.be.false
    })
  })

  describe('even', () => {
    it('should return true for even number', () => {
      expect(_.even(2)).to.be.true
    })

    it('should return false for odd number', () => {
      expect(_.even(1)).to.be.false
    })
  })

  describe('ucf', () => {
    it('should capitalize the first letter', () => {
      expect(_.ucf('hello')).to.equal('Hello')
    })
  })

  describe('uc', () => {
    it('should uppercase the string', () => {
      expect(_.uc('hello')).to.equal('HELLO')
    })
  })

  describe('lc', () => {
    it('should lowercase the string', () => {
      expect(_.lc('HELLO')).to.equal('hello')
    })
  })

  describe('pad', () => {
    it('should pad a string to a certain length', () => {
      expect(_.pad('Hello', 10)).to.equal('     Hello')
    })
  })

  describe('unpad', () => {
    it('should remove leading white spaces from a string', () => {
      expect(_.unpad('     Hello')).to.equal('Hello')
    })
  })

  describe('sleep', () => {
    it('should sleep for specified time in milliseconds', async () => {
      const start = Date.now()
      await _.sleep(100)
      const end = Date.now()
      expect(end - start).to.be.closeTo(100, 10)
    })
  })

  describe('objProp', () => {
    it('should define a property on an object', () => {
      const obj = {}
      _.objProp(obj, 'test', 'value', { enumerable: true })
      expect(obj.test).to.equal('value')
    })
  })

  describe('str2ms', () => {
    it('should convert string duration to milliseconds', () => {
      expect(_.str2ms('1s')).to.equal(1000)
      expect(_.str2ms('1m')).to.equal(60000)
      expect(_.str2ms('1h')).to.equal(3600000)
    })
  })

  describe('sortChars', () => {
    it('should sort characters in a string', () => {
      expect(_.sortChars('hello')).to.equal('ehllo')
    })
  })

  describe('proxyGet', () => {
    it('should create a proxy with custom getter', () => {
      const proxy = _.proxyGet({a: 1}, (target, prop) => target[prop] ? target[prop] : 'not found')
      expect(proxy.a).to.equal(1)
      expect(proxy.b).to.equal('not found')
    })
  })

  describe('proxyNew', () => {
    it('should create a new proxy instance', () => {
      function MyFunc() {}
      const proxy = _.proxyNew(MyFunc)
      expect(proxy()).to.be.instanceOf(MyFunc)
    })
  })

  describe('getMethods', () => {
    it('should get methods of an object', () => {
      const obj = {
        method1: () => {},
        method2: () => {},
        prop: 'value'
      }
      expect(_.getMethods(obj)).to.deep.equal(['method1', 'method2'])
    })
  })

  describe('respond', () => {
    it('should resolve promises and return values', async () => {
      const promise = Promise.resolve('value')
      const result = await _.respond(promise)
      expect(result).to.equal('value')
    })
  })

  describe('encode', () => {
    it('should encode buffer to base58 or hex string', () => {
      const buff = Buffer.from('test')
      expect(_.encode(buff)).to.equal(_.bs58.encode(buff))
      expect(_.encode(buff, true)).to.equal(buff.toString('hex'))
    })
  })

  describe('decode', () => {
    it('should decode base58 or hex string to buffer', () => {
      const str = _.bs58.encode(Buffer.from('test'))
      expect(_.decode(str)).to.deep.equal(Buffer.from('test'))
      expect(_.decode(str, 'hex')).to.deep.equal(Buffer.from(str, 'hex'))
    })
  })
})
