const debugModule = require('../lib/debug')

describe('Debug Module', () => {
  let originalConsoleLog = null

  beforeEach(() => { 
    originalConsoleLog = console.log
  })

  afterEach(() => { 
    console.log = originalConsoleLog
  })

  it('should call console.log when the print function is called', () => { 
    let called = false
    console.log = () => {
      called = true
    }
    debugModule('test')
    expect(called).to.be.true
  })

  it('should call console.log with expected formatted output when the print function is called', () => { 
    let loggedArgs = []
    console.log = (...args) => {
      loggedArgs = args
    }
    debugModule('test')
    expect(loggedArgs[0]).to.include('test')
  })

  it('should not log anything when log level is set to mute', () => { 
    let called = false
    console.log = () => {
      log('<< called >>')
      called = true
    }
    const previousLevel = debugModule.level
    debugModule.level = 'mute'
    debugModule('test')
    expect(called).to.be.false
    debugModule.level = previousLevel
  })

  it('should print error message instead of error object when showErrors is false', () => { 
    let loggedArgs = []
    console.log = (...args) => {
      loggedArgs = args
    }
    const errorMessage = 'test error'
    debugModule.error(new Error(errorMessage))
    expect(loggedArgs[2]).to.include(errorMessage)
    expect(loggedArgs[0]).to.not.include('Error')
  })

  it('should print error object when showErrors is true', () => { 
    const previousLevel = debugModule.level
    debugModule.level = 'debug'
    let loggedArgs = []
    console.log = (...args) => {
      loggedArgs = args
    }
    const showErrors = debugModule.showErrors
    const errorMessage = 'test error'
    debugModule.show('errors', true)
    debugModule.error(new Error(errorMessage))
    expect(loggedArgs[2]).to.be.instanceof(Error)
    debugModule.show('errors', showErrors) // reset the showErrors
    debugModule.level = previousLevel
  })

  ;['debug', 'info', 'warn', 'error'].forEach(level => {
    it(`should call console.log with level set to ${level}`, () => {
      const previousLevel = debugModule.level
      debugModule.level = level
      let loggedArgs = []
      console.log = (...args) => {
        loggedArgs = args
      }
      debugModule[level]('test message')
      expect(loggedArgs[1]).to.include(level)
      expect(loggedArgs[2]).to.include('test message')
      debugModule.level = previousLevel
    })
  })
})
